from getpass import getpass
import simplecrypt

def decrypt(inpath, outpath):
    password = getpass()
    with open(inpath, 'rb') as infile, open(outpath, 'wb') as outfile:
        outfile.write(
            simplecrypt.decrypt(password, infile.read())
            )


def encrypt(inpath, outpath):
    password = getpass()
    with open(inpath, 'rb') as infile, open(outpath, 'wb') as outfile:
        outfile.write(
            simplecrypt.encrypt(password, infile.read())
        )




''' library of functions for correlating services
'''
import numpy as np
from scipy import interpolate


def interpolator(support, times_to_be_evaluated):
    '''
        support : is  a structured array with, 'Time'
                    and a lot of other fields, to be interpolated.

        times_to_be_evaluated : is a 1D array-like holding the times,
                        where the interpolation should be done
    '''
    output_array = np.zeros(len(times_to_be_evaluated), dtype=support.dtype)
    output_array['Time'] = times_to_be_evaluated

    field_names = [name for name in support.dtype.names if 'Time' not in name]
    for name in field_names:
        output_array[name] = interpolate.interp1d(
            support['Time'],
            support[name], axis=0
        )(times_to_be_evaluated)
    return output_array


class CorrelationByInterpolation(object):

    def __init__(self, service_1, service_2, delta_in_seconds):
        self.service_1 = service_1
        self.service_2 = service_2
        self.delta_in_seconds = delta_in_seconds
        self.services = [service_1, service_2]

    def correlate(self):
        ''' correlate by interpolation
        '''
        averaged_times = self.__create_average_times()
        common_averaged_times = self.__common_times(averaged_times)
        new_service_2 = interpolator(self.service_2, common_averaged_times)
        new_service_1 = interpolator(self.service_1, common_averaged_times)
        return new_service_1, new_service_2

    def __common_times(self, averaged_times):
        ''' crop times for interpolation

        times : 1D array, containting times in FACT format
        services: list of services (having a 'Time' field)

        times is cropped such, that times does not contain any time
        ealier than the earliest time in any service, and no time later
        than the latest time in any service.
        '''
        maximal_earliest = max([s['Time'][0] for s in self.services])
        minimal_latest = min([s['Time'][-1] for s in self.services])
        start, stop = np.searchsorted(
            averaged_times,
            [maximal_earliest, minimal_latest])
        return averaged_times[start:stop]

    def __create_average_times(self):
        ''' create 'near' times from the 'Time' fields self.services

        near times, are time stampfs which are nearer to each other than delta
        each group of times, which is near enough creates one averaged timestamp
        which is simply the average of the intire group of times
        '''

        list_of_averages = self.__make_average_bad_name(
            self.service_1['Time'],
            self.service_2['Time'])
        list_of_averages += self.__make_average_bad_name(
            self.service_2['Time'],
            self.service_1['Time'])
        return np.array(list_of_averages)

    def __make_average_bad_name(self, time1, time2):
        ''' helper for the fucntion above. this one is not easy to explain
            maybe this means it needs refactoring
        '''
        # convert times to seconds
        time1 *= 24 * 3600
        time2 *= 24 * 3600
        list_of_averages = []
        left_indices = np.searchsorted(
            time1,
            time2 - self.delta_in_seconds)
        right_indices = np.searchsorted(
            time1,
            time2 + self.delta_in_seconds)

        for index_in_2, tup in enumerate(zip(left_indices, right_indices)):
            left, right = tup
            this_time2 = time2[index_in_2]
            list_of_averages.append(
                (time1[left:right].sum() + this_time2) / (1 + len(time1[left:right]))
            )
        return list_of_averages


class CorrelationByIdentification(object):

    def __init__(self, service_1, service_2, delta_in_seconds):
        if service_1.shape[0] > service_2.shape[0]:
            self.longer = service_1
            self.shorter = service_2
            self.__long_1_short_2 = True
        else:
            self.longer = service_2
            self.shorter = service_1
            self.__long_1_short_2 = False

        self.delta_in_seconds = delta_in_seconds

        # a version of self.longer, which contains as many rows
        # as self.shorter, the timestamps in new_longer are as close to those
        # in self.shorter as possible
        self.__new_longer = None

    def correlate(self):
        ''' make parts of service_1 and service_2 of equal length and timestamps
        '''
        self.__new_longer = self.__identify()
        self.__cut_for_delta()
        return

    def __identify(self):
        ''' return that part of longer, which corresponds to best to shorter
        '''
        t_short = self.shorter['Time']
        t_long = self.longer['Time']
        indices_in_longer = np.searchsorted(a=t_long, v=t_short)
        a_bit_smaller = t_long.take(indices_in_longer - 1, mode='clip')
        a_bit_larger = t_long.take(indices_in_longer, mode='clip')
        indices_nearest = np.where(
            (t_short - a_bit_smaller) < (a_bit_larger - t_short),
            indices_in_longer - 1, indices_in_longer
        )
        return self.longer.take(indices_nearest, mode='clip')

    def __cut_for_delta(self):
        ''' crop services of equal length, according to timestamps
        '''
        time1 = self.shorter['Time'] * 24 * 3600
        time2 = self.__new_longer['Time'] * 24 * 3600
        good = np.abs(time1 - time2) < self.delta_in_seconds

        if self.__long_1_short_2:
            return self.__new_longer[good], self.shorter[good]
        else:
            return self.shorter[good], self.__new_longer[good]



from .camera import (
    get_pixel_coords,
    get_pixel_dataframe,
    camera_distance_mm_to_deg,
)
from . import trigger

__all__ = [
    'get_pixel_coords', 'get_pixel_dataframe', 'camera_distance_mm_to_deg'
]



# coding: utf-8
import pandas as pd
from fact.analysis import calc_theta_camera, calc_theta_offs_camera
from fact.io import read_h5py_chunked
from fact.io import create_empty_h5py_dataset, append_to_h5py_dataset
from fact.instrument.constants import LOCATION
from astropy.coordinates import SkyCoord, AltAz
from astropy.time import Time
from joblib import Parallel, delayed
import h5py
import click


def calc_theta_source(df, source):
    obstime = Time(pd.to_datetime(df['timestamp']).dt.to_pydatetime())

    altaz = AltAz(location=LOCATION, obstime=obstime)
    source_altaz = source.transform_to(altaz)

    df['theta_deg'] = calc_theta_camera(
        df.source_x_prediction,
        df.source_y_prediction,
        source_altaz.zen.deg,
        source_altaz.az.deg,
        zd_pointing=df['pointing_position_zd'],
        az_pointing=df['pointing_position_az'],
    )
    theta_offs = calc_theta_offs_camera(
        df.source_x_prediction,
        df.source_y_prediction,
        source_altaz.zen.deg,
        source_altaz.az.deg,
        zd_pointing=df['pointing_position_zd'],
        az_pointing=df['pointing_position_az'],
        n_off=5,
    )

    for i, theta_off in enumerate(theta_offs, start=1):
        df['theta_deg_off_{}'.format(i)] = theta_off

    return df


def calc_theta_coordinates(df):

    df['theta_deg'] = calc_theta_camera(
        df.source_x_prediction,
        df.source_y_prediction,
        df['source_position_zd'],
        df['source_position_az'],
        zd_pointing=df['pointing_position_zd'],
        az_pointing=df['pointing_position_az'],
    )
    theta_offs = calc_theta_offs_camera(
        df.source_x_prediction,
        df.source_y_prediction,
        df['source_position_zd'],
        df['source_position_az'],
        zd_pointing=df['pointing_position_zd'],
        az_pointing=df['pointing_position_az'],
        n_off=5,
    )

    for i, theta_off in enumerate(theta_offs, start=1):
        df['theta_deg_off_{}'.format(i)] = theta_off

    return df


cols = [
    'theta_deg' if i == 0 else 'theta_deg_off_{}'.format(i)
    for i in range(6)
]


@click.command()
@click.argument('INPUTFILE')
@click.option(
    '-s', '--source',
    help='Source name, if not given, take `source_position_az`, and `source_position_zd`'
)
@click.option('-c', '--chunksize', type=int, default=10000)
@click.option('-y', '--yes', is_flag=True, help='Do not ask to overwrite existing keys')
def main(inputfile, source, chunksize, yes):
    '''
    Calculate theta_deg and theta_deg_offs from source position in camera coordinates
    e.g. for example for files analysed with the classifier-tools

    The following keys have to be present in the h5py hdf5 file.
        * az_tracking
        * zd_tracking
        * source_x_prediction
        * source_y_prediction
        * unix_time_utc (Only if a source name is given)
    '''

    with h5py.File(inputfile, 'r') as f:
        if any(col in f['events'].keys() for col in cols) and not yes:
            click.confirm('Output keys already exist, overwrite? ', abort=True)

    if source is None:
        df_it = read_h5py_chunked(
            inputfile,
            key='events',
            columns=[
                'pointing_position_az',
                'pointing_position_zd',
                'source_x_prediction',
                'source_y_prediction',
                'source_position_az',
                'source_position_zd',
            ],
            chunksize=chunksize
        )
        with Parallel(-1, verbose=10) as pool:

            dfs = pool(
                delayed(calc_theta_coordinates)(df)
                for df, start, stop in df_it
            )
    else:
        source = SkyCoord.from_name(source)

        df_it = read_h5py_chunked(
            inputfile,
            key='events',
            columns=[
                'pointing_position_az',
                'pointing_position_zd',
                'source_x_prediction',
                'source_y_prediction',
                'timestamp',
            ],
            chunksize=chunksize
        )

        with Parallel(-1, verbose=10) as pool:

            dfs = pool(
                delayed(calc_theta_source)(df, source)
                for df, start, stop in df_it
            )

    df = pd.concat(dfs)

    with h5py.File(inputfile, mode='r+') as f:
        for i in range(6):
            if i == 0:
                col = 'theta_deg'
            else:
                col = 'theta_deg_off_{}'.format(i)

            if col in f['events']:
                del f['events'][col]
            create_empty_h5py_dataset(df[col].values, f['events'], col)
            append_to_h5py_dataset(df[col].values, f['events'][col])


if __name__ == '__main__':
    main()



from matplotlib.patches import RegularPolygon
from matplotlib.collections import PatchCollection

import numpy as np
import matplotlib.pyplot as plt

from ..instrument import get_pixel_coords
from .utils import calc_linewidth, calc_text_size

lastpixel = -1


def onpick(event):
    global lastpixel
    hitpixel = event.ind[0]
    if hitpixel != lastpixel:
        lastpixel = hitpixel
        plot = event.artist

        ecols = plot.get_edgecolors()
        before = np.array(ecols[hitpixel])
        ecols[hitpixel] = [1, 0, 0, 1]
        plot.set_edgecolors(ecols)
        plt.draw()
        ecols[hitpixel] = before

        print('chid:', hitpixel)
        print('value', plot.get_array()[hitpixel])


def camera(
        data,
        ax=None,
        cmap='gray',
        vmin=None,
        vmax=None,
        pixelcoords=None,
        edgecolor='k',
        linewidth=None,
        picker=False,
        ):
    '''
    Parameters
    ----------

    data : array like with shape 1440
        the data you want to plot into the pixels
    ax : a matplotlib.axes.Axes instace or None
        The matplotlib axes in which to plot. If None, plt.gca() is used
    cmap : str or matplotlib colormap instance
        the colormap to use for plotting the 'dataset'
        [default: gray]
    vmin : float
        the minimum for the colorbar, if None min(data) is used
        [default: None]
    vmax : float
        the maximum for the colorbar, if None max(data) is used
        [default: None]
    pixelcoords : the coordinates for the pixels in form [x-values, y-values]
        if None, the package resource is used
        [default: None]
    edgecolor : any matplotlib color
        the color around the pixel
    picker: bool
        if True then the the pixel are made clickable to show information
    '''

    if ax is None:
        ax = plt.gca()

    ax.set_aspect('equal')

    if picker is True:
        fig = ax.get_figure()
        fig.canvas.mpl_connect('pick_event', onpick)

    # if the axes limit is still (0,1) assume new axes
    if ax.get_xlim() == (0, 1) and ax.get_ylim() == (0, 1):
        ax.set_xlim(-200, 200)
        ax.set_ylim(-200, 200)

    if pixelcoords is None:
        pixel_x, pixel_y = get_pixel_coords()
    else:
        pixel_x, pixel_y = pixelcoords

    if vmin is None:
        vmin = np.min(data)
    if vmax is None:
        vmax = np.max(data)

    edgecolors = np.array(1440 * [edgecolor])
    patches = []
    for x, y, ec in zip(pixel_x, pixel_y, edgecolors):
        patches.append(
            RegularPolygon(
                xy=(x, y),
                numVertices=6,
                radius=9.51/np.sqrt(3),
                orientation=0.,   # in radians
            )
        )

    if linewidth is None:
        linewidth = calc_linewidth(ax=ax)

    collection = PatchCollection(patches, picker=0)
    collection.set_linewidth(linewidth)
    collection.set_edgecolors(edgecolors)
    collection.set_cmap(cmap)
    collection.set_array(data)
    collection.set_clim(vmin, vmax)

    ax.add_collection(collection)
    plt.draw_if_interactive()
    return collection


def pixelids(ax=None, size=None, pixelcoords=None, *args, **kwargs):
    '''
    plot the chids into the pixels
    '''
    if pixelcoords is None:
        pixel_x, pixel_y = get_pixel_coords()
    else:
        pixel_x, pixel_y = pixelcoords

    if ax is None:
        ax = plt.gca()

    if size is None:
        size = calc_text_size(ax)

    x1, x2 = ax.get_xlim()
    y1, y2 = ax.get_ylim()

    maskx = np.logical_and(pixel_x + 4.5 < x2, pixel_x - 4.5 > x1)
    masky = np.logical_and(pixel_y + 4.5 < y2, pixel_y - 4.5 > y1)
    mask = np.logical_and(maskx, masky)

    chids = np.arange(1440)
    for px, py, chid in zip(pixel_x[mask], pixel_y[mask], chids[mask]):
        ax.text(
            px, py,
            str(chid),
            size=size,
            va='center',
            ha='center',
            **kwargs
        )

    plt.draw_if_interactive()


def mark_pixel(pixels, color='g', ax=None, linewidth=None):
    ''' surrounds pixels given by pixels with a border '''
    pixel_x, pixel_y = get_pixel_coords()

    if ax is None:
        ax = plt.gca()

    patches = []
    for xy in zip(pixel_x[pixels], pixel_y[pixels]):
        patches.append(
            RegularPolygon(
                xy=xy,
                numVertices=6,
                radius=9.5 / np.sqrt(3),
                orientation=0.,   # in radians
                fill=False,
            )
        )

    if linewidth is None:
        linewidth = calc_linewidth(ax=ax)

    collection = PatchCollection(patches, picker=0)
    collection.set_linewidth(linewidth)
    collection.set_edgecolors(color)
    collection.set_facecolor('none')

    ax.add_collection(collection)

    plt.draw_if_interactive()
    return collection



from astropy.coordinates import (
    BaseCoordinateFrame,
    AltAz,
    frame_transform_graph,
    FunctionTransform,
)

from astropy.coordinates import (
    CoordinateAttribute,
    TimeAttribute,
    EarthLocationAttribute,
    Attribute
)

from astropy.coordinates.matrix_utilities import rotation_matrix
from astropy.coordinates.representation import CartesianRepresentation
import astropy.units as u

from .representation import PlanarRepresentation
from ..instrument.constants import FOCAL_LENGTH_MM, LOCATION
import numpy as np

focal_length = FOCAL_LENGTH_MM * u.mm


class CameraFrame(BaseCoordinateFrame):
    '''
    Astropy CoordinateFrame representing coordinates in the CameraPlane

    Attributes
    ----------
    pointing_direction: astropy.coordinates.AltAz
        The pointing direction of the telescope
    obstime: astropy.Time
        The timestamp of the observation, only needed to directly transform
        to Equatorial coordinates, transforming to AltAz does not need this.
    location: astropy.coordinates.EarthLocation
        The location of the observer,  only needed to directly transform
        to Equatorial coordinates, transforming to AltAz does not need this,
        default is FACT's location
    rotated: bool
        True means x points right and y points up when looking on the camera
        from the dish, which is the efinition of FACT-Tools >= 1.0 and Mars.
        False means x points up and y points left,
        which is definition in the original FACTPixelMap file.
    '''
    default_representation = PlanarRepresentation
    pointing_direction = CoordinateAttribute(frame=AltAz, default=None)
    obstime = TimeAttribute(default=None)
    location = EarthLocationAttribute(default=LOCATION)
    rotated = Attribute(default=True)


@frame_transform_graph.transform(FunctionTransform, CameraFrame, AltAz)
def camera_to_altaz(camera, altaz):
    if camera.pointing_direction is None:
        raise ValueError('Pointing Direction must not be None')

    x = camera.x.copy()
    y = camera.y.copy()

    if camera.rotated is True:
        x, y = y, -x

    z = 1 / np.sqrt(1 + (x / focal_length)**2 + (y / focal_length)**2)
    x *= z / focal_length
    y *= z / focal_length

    cartesian = CartesianRepresentation(x, y, z, copy=False)

    rot_z_az = rotation_matrix(-camera.pointing_direction.az, 'z')
    rot_y_zd = rotation_matrix(-camera.pointing_direction.zen, 'y')

    cartesian = cartesian.transform(rot_y_zd)
    cartesian = cartesian.transform(rot_z_az)

    altitude = 90 * u.deg - np.arccos(cartesian.z)
    azimuth = np.arctan2(cartesian.y, cartesian.x)

    return AltAz(
        alt=altitude,
        az=azimuth,
        location=camera.location,
        obstime=camera.obstime,
    )


@frame_transform_graph.transform(FunctionTransform, AltAz, CameraFrame)
def altaz_to_camera(altaz, camera):
    if camera.pointing_direction is None:
        raise ValueError('Pointing Direction must not be None')

    cartesian = altaz.cartesian

    rot_z_az = rotation_matrix(camera.pointing_direction.az, 'z')
    rot_y_zd = rotation_matrix(camera.pointing_direction.zen, 'y')

    cartesian = cartesian.transform(rot_z_az)
    cartesian = cartesian.transform(rot_y_zd)

    x = (cartesian.x * focal_length / cartesian.z).copy()
    y = (cartesian.y * focal_length / cartesian.z).copy()

    if camera.rotated is True:
        x, y = -y, x

    return CameraFrame(
        x=x,
        y=y,
        pointing_direction=camera.pointing_direction,
        obstime=altaz.obstime,
        location=camera.location,
    )



from .statistics import li_ma_significance
from .binning import ontime_binning, qla_binning, groupby_observation_blocks, bin_runs

from .core import calc_run_summary_source_independent, split_on_off_source_independent

from .source import (
    calc_theta_equatorial,
    calc_theta_camera,
    calc_theta_offs_camera,
)


__all__ = [
    'li_ma_significance',
    'ontime_binning',
    'qla_binning',
    'groupby_observation_blocks',
    'bin_runs',
    'calc_run_summary_source_independent',
    'split_on_off_source_independent',
    'calc_theta_equatorial',
    'calc_theta_camera',
    'calc_theta_offs_camera',
]
